# -*- coding: utf-8 -*-

"""HelpFormatter

We have a slightly adjusted HelpFormatter that we use for the manpages.

Author: G.J.J. van den Burg
License: See the LICENSE file.
Copyright: 2021, G.J.J. van den Burg

This file is part of Wilderness.
"""

import argparse
import re
import textwrap

from typing import Dict
from typing import List
from typing import Optional


class HelpFormatter(argparse.HelpFormatter):
    def _fill_text(self, text, width, indent):
        # Minor change to _fill_text to keep newlines provided by the user in
        # the prolog and epilog.
        lines = text.splitlines()
        new_lines = []
        for line in lines:
            new_line = textwrap.fill(
                line,
                width,
                initial_indent=indent,
                subsequent_indent=indent,
            )
            new_lines.append(new_line)
        new_text = "\n".join(new_lines)
        return new_text

    def _format_actions_usage(self, actions, groups, return_parts=False):
        # find group indices and identify actions in groups
        group_actions = set()
        inserts = {}  # type: Dict[int, str]
        for group in groups:
            try:
                start = actions.index(group._group_actions[0])
            except ValueError:
                continue
            else:
                end = start + len(group._group_actions)
                if actions[start:end] == group._group_actions:
                    for action in group._group_actions:
                        group_actions.add(action)
                    if not group.required:
                        if start in inserts:
                            inserts[start] += " ["
                        else:
                            inserts[start] = "["
                        if end in inserts:
                            inserts[end] += "]"
                        else:
                            inserts[end] = "]"
                    else:
                        if start in inserts:
                            inserts[start] += " ("
                        else:
                            inserts[start] = "("
                        if end in inserts:
                            inserts[end] += ")"
                        else:
                            inserts[end] = ")"
                    for i in range(start + 1, end):
                        inserts[i] = "|"

        # collect all actions format strings
        parts = []  # type: List[Optional[str]]
        for i, action in enumerate(actions):
            if isinstance(action, argparse._HelpAction):
                inserts = {i - 1: inserts[i] for i in inserts}

            elif action.help is argparse.SUPPRESS:
                parts.append(None)
                if inserts.pop(i) == "|":
                    inserts.pop(i)
                elif inserts.get(i + 1) == "|":
                    inserts.pop(i + 1)

            # produce all arg strings
            elif not action.option_strings:
                default = self._get_default_metavar_for_positional(action)
                part = self._format_args(action, default)

                # if it's in a group, strip the outer []
                if action in group_actions:
                    if part[0] == "[" and part[-1] == "]":
                        part = part[1:-1]
                else:
                    if action.nargs == "?":
                        pass
                    else:
                        part = "<%s>" % part

                # add the action string to the list
                parts.append(part)

            # produce the first way to invoke the option in brackets
            else:
                part = self._format_part(action, group_actions)

                # add the action string to the list
                parts.append(part)

        # insert things at the necessary indices
        for i in sorted(inserts, reverse=True):
            parts[i:i] = [inserts[i]]

        # join all the action items with spaces
        text = " ".join([item for item in parts if item is not None])

        # clean up separators for mutually exclusive groups
        open = r"[\[(]"
        close = r"[\])]"
        text = re.sub(r"(%s) " % open, r"\1", text)
        text = re.sub(r" (%s)" % close, r"\1", text)
        text = re.sub(r"%s *%s" % (open, close), r"", text)
        text = re.sub(r"\(([^|]*)\)", r"\1", text)
        text = text.strip()

        if return_parts:
            return text, parts

        # return the text
        return text

    def _format_part(self, action, group_actions):
        option_string = action.option_strings[0]

        # if the Optional doesn't take a value, format is:
        #    -s or --long
        if action.nargs == 0:
            if len(action.option_strings) == 2:
                part = " | ".join(action.option_strings)
            else:
                part = action.option_strings[0]
        elif option_string.startswith("--"):  # TODO: get prefix from parser
            default = self._get_default_metavar_for_optional(action)
            args_string = self._format_args(action, default)
            if args_string.startswith("["):
                part = "%s[=%s" % (option_string, args_string[1:])
            else:
                part = "%s=%s" % (option_string, args_string)
        elif len(action.option_strings) == 2:
            default = self._get_default_metavar_for_optional(action)
            args_string = self._format_args(action, default)
            part0 = "%s %s" % (action.option_strings[0], args_string)
            part1 = "%s=%s" % (action.option_strings[1], args_string)
            part = "%s | %s" % (part0, part1)

        # if the Optional takes a value, format is:
        #    -s ARGS or --long ARGS
        else:
            default = self._get_default_metavar_for_optional(action)
            args_string = self._format_args(action, default)
            part = "%s %s" % (option_string, args_string)

        # make it look optional if it's not required or in a group
        if not action.required and action not in group_actions:
            part = "[%s]" % part

        return part



# -*- coding: utf-8 -*-

from .application import Application
from .command import Command
from .group import Group
from .manpages import build_manpages
from .tester import Tester

__all__ = [
    "Application",
    "Command",
    "Tester",
    "Group",
    "build_manpages",
]



# -*- coding: utf-8 -*-

"""DocumentableMixin definitions

A documentable is either an application or command, for which we can generate a 
manpage.

Author: G.J.J. van den Burg
License: See the LICENSE file.
Copyright: 2021, G.J.J. van den Burg

This file is part of Wilderness.
"""

import abc
import argparse

from typing import Dict
from typing import Optional

from .formatter import HelpFormatter
from .manpages import ManPage


class DocumentableMixin(metaclass=abc.ABCMeta):
    def __init__(
        self,
        description: Optional[str] = None,
        extra_sections: Optional[Dict[str, str]] = None,
        options_prolog: Optional[str] = None,
        options_epilog: Optional[str] = None,
    ):
        self._description = description  # type: Optional[str]
        self._args = None  # type: Optional[argparse.Namespace]
        self._parser = None  # type: Optional[argparse.ArgumentParser]
        self._arg_help = {}  # type: Dict[str, Optional[str]]
        self._extra_sections = {} if extra_sections is None else extra_sections
        self._options_extra = {
            "prolog": options_prolog or "",
            "epilog": options_epilog or "",
        }

    @property
    def description(self) -> Optional[str]:
        return self._description

    @property
    def parser(self) -> argparse.ArgumentParser:
        assert self._parser is not None
        return self._parser

    @parser.setter
    def parser(self, parser: argparse.ArgumentParser):
        self._parser = parser

    @property
    def args(self) -> argparse.Namespace:
        """The parsed command line arguments"""
        assert self._args is not None
        return self._args

    @args.setter
    def args(self, args: argparse.Namespace):
        self._args = args

    @property
    def argument_help(self) -> Dict[str, Optional[str]]:
        return self._arg_help

    def get_synopsis(self, width: int = 80) -> str:
        optionals = []
        positionals = []
        for action in self.parser._actions:
            if action.option_strings:
                optionals.append(action)
            else:
                positionals.append(action)

        helpfmt = HelpFormatter(prog=self.parser.prog)
        format = helpfmt._format_actions_usage
        _, parts = format(
            optionals + positionals,
            self.parser._mutually_exclusive_groups,
            return_parts=True,
        )

        text = ""
        line = self.parser.prog
        lead = len(line) + 1
        for item in parts:
            if item is None:
                continue
            if len(line) + 1 + len(item) <= width:
                line += " " + item
            else:
                text += line + "\n"
                line = " " * lead + item
        text += line
        return text

    def get_options_text(self) -> str:
        prolog = self._options_extra["prolog"]
        if not isinstance(prolog, str):
            raise ValueError(
                f"Options prolog not of type str (received: {prolog})"
            )
        epilog = self._options_extra["epilog"]
        if not isinstance(epilog, str):
            raise ValueError(
                f"Options epilog not of type str (received: {epilog})"
            )

        text = [prolog]
        text.append("")
        for action in self.parser._get_optional_actions():
            desc = self._arg_help.get(action.dest)
            desc_source = "description"
            if desc is None:
                desc = action.help
                desc_source = "help"

            if desc is argparse.SUPPRESS or desc is None:
                continue

            if not isinstance(desc, str):
                raise ValueError(
                    f"Description for action {action.dest} found in "
                    f"{desc_source} attribute is not of type str "
                    f"(received type: {type(desc)})."
                )

            # TODO clean this up and make it testable
            if action.metavar is None:
                opts = ", ".join(action.option_strings)
            else:
                if action.option_strings[0].startswith(
                    2 * self.parser.prefix_chars
                ):
                    u = action.option_strings[0]
                    if action.choices and action.default:
                        v = f"[=({'|'.join(action.choices)})]"
                    elif action.choices:
                        v = f"=({'|'.join(action.choices)})"
                    else:
                        if action.nargs is None:
                            v = f"={action.metavar}"
                        elif action.nargs == "?":
                            v = f"[={action.metavar}]"
                        else:
                            v = f"={action.metavar}"
                    opts = f"{u}{v}"
                else:
                    opts = f"{action.option_strings[0]} {action.metavar}"
                    if len(action.option_strings) > 1:
                        opts += (
                            f", {action.option_strings[1]}={action.metavar}"
                        )

            text.append(opts)
            text.append(".RS 4")
            text.append(desc)
            text.append(".RE")
            text.append(".PP")

        for action in self.parser._get_positional_actions():
            desc = self._arg_help.get(action.dest)
            desc_source = "description"
            if desc is None:
                desc = action.help
                desc_source = "help"

            if desc is argparse.SUPPRESS or desc is None:
                continue

            if not isinstance(desc, str):
                raise ValueError(
                    f"Description for action {action.dest} found in "
                    f"{desc_source} attribute is not of type str "
                    f"(received type: {type(desc)})."
                )

            if action.nargs == "?":
                text.append(f"[{action.dest}]")
            else:
                text.append(f"<{action.dest}>")
            text.append(".RS 4")
            text.append(desc)
            text.append(".RE")
            text.append(".PP")

        text.append(epilog)
        return "\n".join(text)

    @abc.abstractmethod
    def create_manpage(self) -> ManPage:
        pass



# -*- coding: utf-8 -*-

"""Group definitions

This module contains the definitions for the Group class, which is used to 
collect distinct Command objects for the application.

Author: G.J.J. van den Burg
License: See the LICENSE file.
Copyright: 2021, G.J.J. van den Burg

This file is part of Wilderness.

"""

import argparse

from typing import TYPE_CHECKING
from typing import Dict
from typing import List
from typing import Optional

if TYPE_CHECKING:
    from wilderness import Application
    from wilderness import Command


class Group:
    def __init__(self, title: Optional[str] = None, is_root: bool = False):
        self._title = title
        self._is_root = is_root

        self._command_map = {}  # type: Dict[str, Command]
        self._app = None  # type: Optional[Application]

    @property
    def application(self) -> Optional["Application"]:
        return self._app

    @property
    def title(self) -> Optional[str]:
        return self._title

    @property
    def commands(self) -> List["Command"]:
        return list(self._command_map.values())

    @property
    def is_root(self) -> bool:
        """Return whether the groups is its Application's root group"""
        return self._is_root

    def commands_as_actions(self) -> List[argparse.Action]:
        actions = []
        for command in self.commands:
            action = argparse.Action(
                option_strings=[], dest=command.name, help=command.title
            )
            actions.append(action)
        return actions

    def set_app(self, app: "Application") -> None:
        self._app = app

    def add(self, command: "Command") -> None:
        self._command_map[command.name] = command
        assert self.application is not None
        self.application._add_command(command)

    def __len__(self) -> int:
        return len(self.commands)



# -*- coding: utf-8 -*-

"""ArgumentParser override

Author: G.J.J. van den Burg
License: See the LICENSE file.
Copyright: 2021, G.J.J. van den Burg

This file is part of Wilderness.

"""

import argparse
import sys

from typing import TYPE_CHECKING
from typing import Optional

if TYPE_CHECKING:
    from .command import Command


class ArgumentParser(argparse.ArgumentParser):
    def __init__(self, *args, exit_on_error=True, **kwargs):
        super().__init__(*args, **kwargs)
        self.exit_on_error = exit_on_error
        self._exit_called = False

    def exit(self, status: Optional[int] = 0, message: Optional[str] = None):
        if message:
            self._print_message(message, sys.stderr)
        self._exit_called = True
        if self.exit_on_error:
            sys.exit(status)


class ArgumentGroup:
    def __init__(self, group: argparse._ArgumentGroup):
        self._group = group
        self._command = None  # type: Optional[Command]

    @property
    def command(self) -> Optional["Command"]:
        return self._command

    @command.setter
    def command(self, command: "Command"):
        self._command = command

    def add_argument(self, *args, **kwargs):
        assert not self.command is None
        description = kwargs.pop("description", None)
        action = self._group.add_argument(*args, **kwargs)
        self.command.argument_help[action.dest] = description
        return action


class MutuallyExclusiveGroup:
    def __init__(self, meg: argparse._MutuallyExclusiveGroup):
        self._meg = meg
        self._command = None  # type: Optional[Command]

    @property
    def command(self) -> Optional["Command"]:
        return self._command

    @command.setter
    def command(self, command: "Command"):
        self._command = command

    def add_argument(self, *args, **kwargs):
        assert not self.command is None
        description = kwargs.pop("description", None)
        action = self._meg.add_argument(*args, **kwargs)
        self.command.argument_help[action.dest] = description
        return action



# -*- coding: utf-8 -*-

"""Help command definitions

This module contains the definitions for our HelpCommand and our HelpAction.  
The HelpCommand takes care of opening the manpage when the "help" subcommand is 
called, and the HelpAction is slightly modified to use our help text formatter 
(see the Application class).

Author: G.J.J. van den Burg
License: See the LICENSE file.
Copyright: 2021, G.J.J. van den Burg

This file is part of Wilderness.
"""

import argparse
import subprocess
import sys

from typing import TYPE_CHECKING

from .command import Command

if TYPE_CHECKING:
    from .application import Application


class HelpCommand(Command):
    def __init__(self):
        super().__init__(
            name="help",
            title="Display help information",
            description="Display help information",
        )

    def handle(self) -> int:
        assert self.args is not None
        assert self.application

        cmd = self.args.command
        if cmd is None:
            self.application.print_help()
            return 1

        if not have_man_command():
            print("Error: man command not available.", file=sys.stderr)
            return 2

        app_name = self.application.name
        cp = subprocess.run(["man", f"{app_name}-{cmd}"])
        return cp.returncode

    def register(self):
        self.add_argument(
            "command",
            nargs="?",
            # help=argparse.SUPPRESS,
            description=argparse.SUPPRESS,
        )


def help_action_factory(app: "Application"):
    class HelpAction(argparse.Action):
        _app = app

        def __init__(
            self,
            option_strings,
            dest=argparse.SUPPRESS,
            default=argparse.SUPPRESS,
            help=None,
        ):
            super().__init__(
                option_strings=option_strings,
                dest=dest,
                default=default,
                nargs=0,
                help=help,
            )

        def __call__(self, parser, namespace, values, option_string=None):
            if self._app is None:
                parser.print_help()
            else:
                self._app.print_help()
            parser.exit()

    return HelpAction


def have_man_command() -> bool:
    try:
        cp = subprocess.run(
            ["man", "--version"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            check=False,
        )
    except FileNotFoundError:
        return False
    return cp.returncode == 0



# -*- coding: utf-8 -*-

"""Tester class

This module contains the CommandTester class.

Author: G.J.J. van den Burg
License: See the LICENSE file.
Copyright: 2021, G.J.J. van den Burg

This file is part of Wilderness.
"""

import contextlib
import io

from typing import List
from typing import Optional

from .application import Application


class Tester:
    def __init__(self, app: Application) -> None:
        self._app = app

    @property
    def application(self) -> Application:
        return self._app

    def clear(self):
        self._retcode = None
        self._io_stderr = None
        self._io_stdout = None

    def get_return_code(self) -> Optional[int]:
        return self._retcode

    def get_stdout(self) -> Optional[str]:
        if self._io_stdout is None:
            return None
        return self._io_stdout.getvalue()

    def get_stderr(self) -> Optional[str]:
        if self._io_stderr is None:
            return None
        return self._io_stderr.getvalue()

    def test_command(self, cmd_name: str, args: List[str]) -> None:
        self.clear()
        command = self.application.get_command(cmd_name)
        if command is None:
            raise ValueError(f"No such command: {cmd_name}")

        args.insert(0, cmd_name)
        parser = self.application._parser
        parser.exit_on_error = False
        parsed_args = parser.parse_args(args=args)
        command.args = parsed_args

        self._io_stdout = io.StringIO()
        self._io_stderr = io.StringIO()

        with contextlib.redirect_stdout(self._io_stdout):
            with contextlib.redirect_stderr(self._io_stderr):
                self._retcode = command.handle()

    def test_application(self, args: Optional[List[str]] = None) -> None:
        self.clear()
        args = [] if args is None else args

        self._io_stdout = io.StringIO()
        self._io_stderr = io.StringIO()

        with contextlib.redirect_stdout(self._io_stdout):
            with contextlib.redirect_stderr(self._io_stderr):
                self._retcode = self.application.run(
                    args=args, exit_on_error=False
                )



# -*- coding: utf-8 -*-

"""Application class

This module contains the Application class.

Author: G.J.J. van den Burg
License: See the LICENSE file.
Copyright: 2021, G.J.J. van den Burg

This file is part of Wilderness.
"""

import argparse
import sys

from typing import Dict
from typing import List
from typing import Optional
from typing import TextIO

from ._argparse import ArgumentParser
from .command import Command
from .documentable import DocumentableMixin
from .formatter import HelpFormatter
from .group import Group
from .help import HelpCommand
from .help import help_action_factory
from .manpages import ManPage


class Application(DocumentableMixin):
    """Base class for applications

    This is the main Application object that Wilderness applications are
    expected to inherit from. All text that is supplied to the man pages, such
    as the description, can use basic formatting constructs documented in the
    :obj:`ManPage.groffify() <wilderness.manpages.ManPage.groffify>` method.

    Parameters
    ----------
    name : str
        The name of the application.

    version : str
        The version of the application, to be used in creating the man pages.

    author : Optional[str]
        The author(s) of the application. This is used in the man pages, but is
        not actually visible in the output (it is recorded in the metadata
        header of the man pages).

    title : Optional[str]
        The title of the application is used as a short description. It shows
        up in the man pages as the text after the application name in the first
        section.

    description : Optional[str]
        Long description of the application. This is used in the man pages in
        the DESCRIPTION section after the synopsis.

    default_command : Optional[str]
        The default command to run when none is supplied on the command line.
        By default this is omitted and the help text is shown instead, but some
        applications may want to run a particular command as default instead.

    add_help : bool
        Whether to add help commands or not. This adds support for the
        traditional help flags ``-h`` or ``--help`` for the short help text on
        the command line, as well as the ``help`` command that opens the man
        pages for the subcommands of the application. Note that the short help
        text on the command line typically provides a list of available
        commands.

        See the `FakeDF`_ example for an application where this is not
        enabled.

    extra_sections : Optional[Dict[str, str]]
        Additional sections of documentation for the man page. This is expected
        to be provided as a dictionary where the keys are the section headers
        and the values are the section text. Basic formatting constructs such
        as lists and enumerations are understood by the text processor (see
        :obj:`ManPage.groffify() <wilderness.manpages.ManPage.groffify>` for
        further details).

    prolog : Optional[str]
        Text to be shown in the short command line help text, before the
        (grouped) list of available commands. Newline characters are preserved.

    epilog: Optional[str]
        Text to be shown in the short command line help text, after the list of
        available commands. Newline characters are preserved.

    options_prolog: Optional[str]
        Text to be shown in the man page before the list of options. See the
        `FakeDF`_ application for an example.

    options_epilog: Optional[str]
        Text to be shown in the man page after the list of options. See the
        `FakeDF`_ application for an example.

    add_commands_section: bool
        Whether to automatically generate a section in the application man page
        that lists the available commands.

    .. _FakeDF:
        https://github.com/GjjvdBurg/wilderness/tree/master/examples/fakedf

    """

    _cmd_name = "command"

    def __init__(
        self,
        name: str,
        version: str,
        author: Optional[str] = None,
        title: Optional[str] = None,
        description: Optional[str] = None,
        default_command: Optional[str] = None,
        add_help: bool = True,
        extra_sections: Optional[Dict[str, str]] = None,
        prolog: Optional[str] = None,
        epilog: Optional[str] = None,
        options_prolog: Optional[str] = None,
        options_epilog: Optional[str] = None,
        add_commands_section: bool = False,
    ):
        super().__init__(
            description=description,
            extra_sections=extra_sections,
            options_prolog=options_prolog,
            options_epilog=options_epilog,
        )

        self._name = name
        self._version = version
        self._author = "" if author is None else author
        self._title = title
        self._default_command = default_command
        self._add_help = add_help

        self._parser = ArgumentParser(
            prog=name,
            description=prolog,
            epilog=epilog,
            formatter_class=HelpFormatter,
            add_help=False,
        )  # type: ArgumentParser
        self._subparsers = None  # type: Optional[argparse._SubParsersAction]

        self._command_map = {}  # type: Dict[str, Command]
        self._group_map = {}  # type: Dict[str, Group]
        self._root_group = None  # type: Optional[Group]
        self._args = None  # type: Optional[argparse.Namespace]

        self._prolog = prolog
        self._epilog = epilog

        self._add_commands_section = add_commands_section

        # TODO: allow the user to set this and extract from self._parser
        default_prefix = "-"
        if self._add_help:
            self._parser.add_argument(
                default_prefix + "h",
                default_prefix * 2 + "help",
                action=help_action_factory(self),
                default=argparse.SUPPRESS,
                help="show this help message and exit",
            )
            self.add(HelpCommand())

        self.register()

    @property
    def name(self) -> str:
        """The name of the application"""
        return self._name

    @property
    def author(self) -> str:
        """The author(s) of the application"""
        return self._author

    @property
    def version(self) -> str:
        """The version of the package or application"""
        return self._version

    @property
    def commands(self) -> List[Command]:
        """List the commands registered to the application

        Returns
        -------
        commands : List[:class:`command.Command`]
            The list of commands registered to the application.

        """
        cmds = []
        if self._root_group:
            cmds.extend(list(self._root_group.commands))
        for group in self._group_map.values():
            cmds.extend(list(group.commands))
        return cmds

    @property
    def groups(self) -> List[Group]:
        """List the groups registered to the application

        If no groups have been added to the application but commands have been
        added, this property will contain a single group, the root group.

        Returns
        -------
        groups: List[:class:`group.Group`]
            The list of groups registered to the application

        """
        groups = []
        if self._root_group is not None:
            groups.append(self._root_group)
        for group in self._group_map.values():
            groups.append(group)
        return groups

    def add_argument(self, *args, **kwargs) -> argparse.Action:
        """Add an argument to the application

        This wraps the argparse.ArgumentParser.add_argument method, with the
        minor difference that it supports a "description" keyword argument,
        which will be used to provide a long help message for the argument in
        the man page.
        """
        help_ = kwargs.get("help", None)
        description = kwargs.pop("description", help_)
        action = self._parser.add_argument(*args, **kwargs)
        self._arg_help[action.dest] = description
        return action

    def add_mutually_exclusive_group(
        self, *args, **kwargs
    ) -> argparse._MutuallyExclusiveGroup:
        """Add a mutually exclusive group

        This wraps the `add_mutually_exclusive_group`_ from the argparse
        module, and uses the same positional and keyword arguments.

        .. _add_mutually_exclusive_group:
            https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser.add_mutually_exclusive_group

        """
        assert self._parser is not None
        group = self._parser.add_mutually_exclusive_group(*args, **kwargs)
        return group

    def add(self, command: Command):
        """Add a command to the application

        Note that the ``register`` method of the command is called when it is
        added to the application.

        Parameters
        ----------
        command : :class:`.command.Command`
            The command to add to the application.

        """
        if self._subparsers is None:
            self._subparsers = self._parser.add_subparsers(
                dest="target", metavar=self._cmd_name
            )
        if self._root_group is None:
            self._root_group = Group(title="Available commands", is_root=True)
            self._root_group.set_app(self)
        self._root_group.add(command)

    def _add_command(self, command: Command):
        assert self._subparsers is not None
        self._command_map[command._name] = command
        cmd_parser = self._subparsers.add_parser(
            command.name,
            help=command.title,
            add_help=command._add_help,
        )
        command.parser = cmd_parser
        command.register()
        command._application = self

    def add_group(self, title: str) -> Group:
        """Create a group of commands

        Parameters
        ----------
        title : str
            The title for the group.

        Returns
        -------
        :class:`.group.Group`
            The created command group.

        """
        group = Group(title)
        group.set_app(self)
        self._group_map[title] = group
        return group

    def register(self):
        """Register arguments to the application

        Override this method to add command line arguments to the application
        (using self.add_argument, etc). For single-command applications, this
        should be used to add all command line arguments. For multi-command
        applications, this method can be used to add arguments that apply to
        all commands, or arguments such as --version.

        This method is called upon initialization of the Application object.

        """
        pass

    def handle(self) -> int:
        """Main method to override for single-command applications.

        When creating a single-command application (such as the `FakeDF`_
        example), this method must be overridden with the actual functionality.
        For multi-command applications, this method is not used.

        Returns
        -------
        return_code : int
            The return code of the application, to be used as the return code
            on the command line.

        """
        return 1

    def run(
        self,
        args: Optional[List[str]] = None,
        namespace: Optional[argparse.Namespace] = None,
        exit_on_error: bool = True,
    ) -> int:
        """Main method to run the application

        Parameters
        ----------
        args : Optional[List[str]]
            List of arguments to the application. This is typically only used
            for testing, as by default the arguments will be read from the
            command line.

        namespace : Optional[argparse.Namespace]
            Namespace object to save the arguments to. By default a new
            argparse.Namespace object is created.

        exit_on_error : bool
            Whether or not to exit when argparse encounters an error.

        Returns
        -------
        return_code : int
            The return code of the application, to be used as the return code
            at the command line.

        """
        # Parse the command line arguments as given
        self._parser.exit_on_error = exit_on_error
        parsed_args = self._parser.parse_args(args=args, namespace=namespace)

        # If a parser error caused argparse to print the help, then we stop
        # here
        if self._parser._exit_called:
            return 1

        # If there are no subparsers registered, we have an application without
        # subcommands, so the application handles things
        self.args = parsed_args
        if self._subparsers is None:
            return self.handle()

        # Satisfy mypy
        assert self.args is not None

        # If we have no target, check if we have a default and set that as the
        # target. If we don't print help and exit.
        if self.args.target is None:
            if self._default_command:
                self.args.target = self._default_command
            else:
                self.print_help()
                return 1

        # Run the requested command
        command = self.get_command(self.args.target)
        command.args = self.args
        return self.run_command(command)

    def run_command(self, command: Command) -> int:
        """Run a particular command directy

        Parameters
        ----------
        command : :class:`.command.Command`
            The command to execute

        Returns
        -------
        return_code : int
            The return code of the handle() method of the command.

        """
        # This is here so the user can override how commands are executed
        return command.handle()

    def get_command(self, command_name: str) -> Command:
        """Get a command by name

        Parameters
        ----------
        command_name : str
            The name of the command to find

        Returns
        -------
        command : :class:`.command.Command`
            The instance of the Command to be returned.

        Raises
        ------
        KeyError
            If no command with the provided name can be found, a KeyError is
            raised.

        """
        return self._command_map[command_name]

    def set_prolog(self, prolog: str) -> None:
        """Set the prolog of the command line help text

        Parameters
        ----------
        prolog : str
            Text to include before the list of commands in the command line
            help text. The prolog is printed after the synopsis of the
            application.

        """
        self._prolog = prolog

    def set_epilog(self, epilog: str) -> None:
        """Set the epilog of the command line help text

        Parameters
        ----------
        epilog : str
            Text to include at the end of the command line help text.

        """
        self._epilog = epilog

    def get_commands_text(self) -> str:
        text = []
        for cmd in self.commands:
            text.append(f"{self.name}-{cmd.name}(1)")
            text.append(f"\t{cmd.title or ''}")
            text.append("")
        return "\n".join(text)

    def create_manpage(self) -> ManPage:
        """Create the Manpage for the application

        Returns
        -------
        man_page : ManPage
            The generated ManPage object.

        """
        man = ManPage(
            self.name,
            version=self._version,
            title=self._title,
            author=self._author,
        )
        man.add_section_synopsis(self.get_synopsis())
        if self.description:
            man.add_section("description", self.description)
        man.add_section("options", self.get_options_text())
        if self._add_commands_section:
            man.add_section("commands", self.get_commands_text())
        for sec in self._extra_sections:
            man.add_section(sec, self._extra_sections[sec])
        return man

    def format_help(self) -> str:
        """Format the command line help for the application

        This method creates the help text for the command line, which is
        typically printed when the -h / --help / help command line arguments
        are used. The :func:`print_help` method calls this method to format
        the help text.

        Returns
        -------
        help_text : str
            The help text as a single string.

        """
        formatter = argparse.RawTextHelpFormatter(prog=self._parser.prog)

        # usage
        formatter.add_usage(
            self._parser.usage,
            self._parser._actions,
            self._parser._mutually_exclusive_groups,
        )

        # prolog
        formatter.add_text(self._prolog)

        # add commands from root group, unless we only have help
        only_help = (
            self._root_group
            and len(self._root_group) == 1
            and self._root_group.commands[0].name == "help"
        )
        if self._root_group and not only_help:
            formatter.start_section(self._root_group.title)
            actions = self._root_group.commands_as_actions()
            formatter.add_arguments(actions)
            formatter.end_section()

        # add commands from other groups
        for group in self._group_map.values():
            formatter.start_section(group.title)
            actions = group.commands_as_actions()
            formatter.add_arguments(actions)
            formatter.end_section()

        # epilog
        formatter.add_text(self._epilog)

        # determine help from format above
        return formatter.format_help()

    def print_help(self, file: TextIO = None):
        """Print the command line help text for the application

        Parameters
        ----------
        file : Optional[TextIO]
            The file to which to write the help text. If omitted, the help text
            will be written to sys.stdout.

        """
        if file is None:
            file = sys.stdout
        message = self.format_help()
        self._parser._print_message(message, file=file)



# -*- coding: utf-8 -*-

"""Code to generate manpages

Author: G.J.J. van den Burg
License: See the LICENSE file.
Copyright: 2021, G.J.J. van den Burg

This file is part of Wilderness.
"""

import datetime as dt
import os
import re

from typing import TYPE_CHECKING
from typing import List
from typing import Optional

if TYPE_CHECKING:
    from wilderness.application import Application


class ManPage:
    def __init__(
        self,
        application_name: str,
        author: Optional[str] = "",
        command_name: Optional[str] = None,
        date: Optional[str] = None,
        title: Optional[str] = None,
        version: Optional[str] = "",
    ):

        self._application_name = application_name
        self._command_name = command_name
        self._version = version

        date = dt.date.today().strftime("%Y-%m-%d") if date is None else date

        self._title = title
        self._metadata = {
            "Title": self.name,
            "Author": author,
            "Generator": "Wilderness <https://pypi.org/project/wilderness>",
            "Date": date,
            "Manual": f"{self._application_name} Manual",
            "Source": f"{self._application_name} {self._version}",
            "Language": "English",
        }

        # Must be last
        self._page = []
        self._page.extend(self.metadata())
        self._page.append(self.header())
        self._page.extend(self.preamble())
        self._page.append(self.section_name())

    @property
    def name(self) -> str:
        app = self._application_name
        cmd = self._command_name
        if cmd is None:
            return app
        return f"{app}-{cmd}"

    def metadata(self) -> List[str]:
        text = ["'\\\" t"]  # This invokes the gtbl preprocessor (unused)

        # Metadata
        maxlen = max(map(len, self._metadata.keys()))
        for key, value in self._metadata.items():
            text.append(f'.\\"{key.rjust(maxlen + 1)}: {value}')
        text.append('.\\"')
        return text

    def preamble(self) -> List[str]:
        text = []

        # Portability stuff
        text.append('.\\" ' + "-" * 65)
        text.append('.\\" * Define some portability stuff')
        text.append('.\\" ' + "-" * 65)
        text.append('.\\" ' + "~" * 65)
        text.append('.\\" http://bugs.debian.org/507673')
        text.append(
            '.\\" http://lists.gnu.org/archive/html/groff/2009-02/msg00013.html'
        )
        text.append('.\\" ' + "~" * 65)
        text.append(".ie \\n(.g .ds Aq \\(aq")
        text.append(".el       .ds Aq '")

        # Default formatting
        text.append('.\\" ' + "-" * 65)
        text.append('.\\" * set default formatting *')
        text.append('.\\" ' + "-" * 65)
        text.append('.\\" disable hyphenation')
        text.append(".nh")
        text.append('.\\" disable justification')
        text.append(".ad l")

        # Banner
        text.append('.\\" ' + "-" * 65)
        text.append('.\\" * MAIN CONTENT STARTS HERE *')
        text.append('.\\" ' + "-" * 65)
        return text

    def header(self) -> str:
        assert isinstance(self._metadata["Date"], str)
        assert isinstance(self._version, str)

        app = self._application_name
        date = self._metadata["Date"].replace("-", "\\-")
        version = self._version.replace(".", "\\&.")
        header = (
            f'.TH "{self.name.upper()}" "1" "{date}" '
            f'"{app.capitalize()} {version}" '
            f'"{app.capitalize()} Manual"'
        )
        return header

    def section_name(self) -> str:
        if self._title is None:
            return f'.SH "NAME"\n{self.name}'
        return f'.SH "NAME"\n{self.name} \\- {self._title}'

    def add_section_synopsis(self, synopsis: str) -> None:
        text = [
            '.SH "SYNOPSIS"',
            ".sp",
            ".nf",
            f"\\fI{self.groffify(synopsis)}",
            ".fi",
            ".sp",
        ]
        self._page.extend(text)

    def add_section(self, label: str, text: str) -> None:
        section = [
            f'.SH "{label.upper()}"',
            ".sp",
            self.groffify(text),
        ]
        self._page.extend(section)

    def groffify(self, text: str) -> str:
        """Format a text line for use in manpages

        This function supports several basic formatting constructs. First,
        newlines in the text are preserved. Next, lists can be created by
        starting lines with :literal:`* \ ` , as long as each list entry starts
        on its own line (that is, separated by :code:`\\n`). Indented text can
        be created by prefixing a line with one or more :code:`\\t` characters.
        Numbered lists are also recognized, as long as each item starts with
        :literal:`1. \ ` (that is, a digit followed by a period, followed by a
        space).

        Parameters
        ----------
        text : str
            The text to convert

        Returns
        -------
        formatted_text : str
            The formatted text ready for use in manpage documents.

        """

        output = []

        lines = text.split("\n")
        for line in lines:
            match = re.match("^\ ?\d+\.\ ", line)
            if line.startswith("* "):
                output.append(".RS 4")
                output.append(".ie n \\{\\")
                output.append("\\h'-04'\\(bu\\h'+03'\\c")
                output.append(".\\}")
                output.append(".el \\{\\")
                output.append(".sp -1")
                output.append(".IP \\(bu 2.3")
                output.append(".\\}")
                output.append(self.groffify_line(line[2:]))
                output.append(".RE")
            elif line.startswith("\t"):
                c = 0
                while line.startswith("\t"):
                    line = line[1:]
                    c += 1
                output.append(f".RS {4 * c}")
                output.append(self.groffify_line(line))
                output.append(".RE")
            elif match:
                label = line[match.start() : match.end()]
                rest = line[match.end() :]
                output.append(f"\\fB{label}\\fR{self.groffify_line(rest)}")
                output.append(".br")
            elif line in ["", "\n"]:
                output.append(".sp")
            else:
                output.append(self.groffify_line(line))
        return "\n".join(output)

    def groffify_line(self, line: str) -> str:
        line = line.replace("\\", "\\e")
        line = line.replace("-", "\\-")
        line = line.replace("...", "\\&...")
        line = line.replace("\n\n", "\n.sp\n")
        return line

    def export(self, output_dir: str) -> str:
        filename = os.path.join(output_dir, f"{self.name}.1")
        with open(filename, "w") as fp:
            fp.write("\n".join(self._page))
        return filename


def build_manpages(app: "Application", output_directory: str = "man") -> None:
    """Write manpages to the output directory

    Parameters
    ----------
    app : Application
        The application for which to generate manpages.

    output_directory : str
        The output directory to which to write the manpages.

    """
    os.makedirs(output_directory, exist_ok=True)
    man = app.create_manpage()
    filename = man.export(output_directory)
    print(f"Wrote manpage to {filename}")
    for cmd in app.commands:
        man = cmd.create_manpage()
        filename = man.export(output_directory)
        print(f"Wrote manpage to {filename}")



# -*- coding: utf-8 -*-

"""Command definition

This module contains the definitions for the Command class.

Author: G.J.J. van den Burg
License: See the LICENSE file.
Copyright: 2021, G.J.J. van den Burg

This file is part of Wilderness.
"""

import abc
import argparse

from typing import TYPE_CHECKING
from typing import Dict
from typing import Optional

from ._argparse import ArgumentGroup
from ._argparse import MutuallyExclusiveGroup
from .documentable import DocumentableMixin
from .manpages import ManPage

if TYPE_CHECKING:
    from .application import Application


class Command(DocumentableMixin, metaclass=abc.ABCMeta):
    def __init__(
        self,
        name: str,
        title: Optional[str] = None,
        description: Optional[str] = None,
        add_help: Optional[bool] = True,
        extra_sections: Optional[Dict[str, str]] = None,
        options_prolog: Optional[str] = None,
        options_epilog: Optional[str] = None,
    ):
        super().__init__(
            description=description,
            extra_sections=extra_sections,
            options_prolog=options_prolog,
            options_epilog=options_epilog,
        )
        self._name = name
        self._title = title

        self._args = None  # type: Optional[argparse.Namespace]
        self._application = None  # type: Optional[Application]
        self._add_help = add_help

    @property
    def application(self) -> Optional["Application"]:
        return self._application

    @property
    def name(self) -> str:
        return self._name

    @property
    def title(self) -> Optional[str]:
        return self._title

    def add_argument(self, *args, **kwargs):
        assert self._parser is not None
        help_ = kwargs.get("help", None)
        description = kwargs.pop("description", help_)
        action = self._parser.add_argument(*args, **kwargs)
        self._arg_help[action.dest] = description
        return action

    def add_argument_group(self, *args, **kwargs) -> ArgumentGroup:
        assert self._parser is not None
        _group = self._parser.add_argument_group(*args, **kwargs)
        group = ArgumentGroup(_group)
        group.command = self
        return group

    def add_mutually_exclusive_group(
        self, *args, **kwargs
    ) -> MutuallyExclusiveGroup:
        assert self._parser is not None
        _meg = self._parser.add_mutually_exclusive_group(*args, **kwargs)
        meg = MutuallyExclusiveGroup(_meg)
        meg.command = self
        return meg

    def register(self):
        pass

    @abc.abstractmethod
    def handle(self) -> int:
        pass

    def create_manpage(self) -> ManPage:
        assert self.application is not None
        man = ManPage(
            self.application.name,
            command_name=self.name,
            version=self.application.version,
            title=self._title,
            author=self.application.author,
        )
        man.add_section_synopsis(self.get_synopsis())
        if self.description:
            man.add_section("description", self.description)
        man.add_section("options", self.get_options_text())
        for sec in self._extra_sections:
            man.add_section(sec, self._extra_sections[sec])
        return man
